import json
import time
import threading
from functools import partial

import psycopg2
import redis

from .ratelimit import RateLimiter
from zope.interface import implements
from twisted.internet import reactor, threads
from twisted.web import server, resource
from txstatsd.itxstatsd import IMetric

ONE_MINUTE = 60
ONE_HOUR = 60 * ONE_MINUTE
ONE_DAY = 24 * ONE_HOUR

MINUTE_LIMITER = RateLimiter(span=ONE_MINUTE, bucket_size=10)
HOUR_LIMITER = RateLimiter(span=ONE_HOUR, bucket_size=10 * ONE_MINUTE)
DAY_LIMITER = RateLimiter(span=ONE_DAY, bucket_size=3 * ONE_HOUR)


class JSONMethodResource(resource.Resource):
    """Renders the result of calling C{name} on C{target} as a json result."""
    isLeaf = True

    def __init__(self, target, name):
        self.target = target
        self.name = name

    def _render_json_error(self, request, result):
        """Fetch the data, make the request return"""
        request.setResponseCode(500)
        request.write(json.dumps({"error": result.getErrorMessage()}))
        request.finish()

    def _render_json_result(self, request, result):
        """Fetch the data, make the request return"""
        request.write(json.dumps({"result": result}))
        request.finish()

    def render_GET(self, request):
        """Return the web request asynchronously."""
        d = threads.deferToThread(
            getattr(self.target, self.name),
            **dict((k, float(v[0])) for k, v in request.args.items()))
        d.addCallback(partial(self._render_json_result, request))
        d.addErrback(partial(self._render_json_error, request))
        return server.NOT_DONE_YET


class DistinctResource(resource.Resource):
    """Returns statistics about unique values."""

    def __init__(self, reporter):
        resource.Resource.__init__(self)
        self.reporter = reporter
        self.putChild("top", JSONMethodResource(self.reporter,
                                                "_get_distinct_top_value"))
        self.putChild("count", JSONMethodResource(self.reporter,
                                                  "_get_distinct_count"))


class DistinctMetricReporter(object):
    """
    Keeps an mesurement of the distinct numbers of items seen and the times
    it has seen each one.
    """
    implements(IMetric)

    periods = (5 * ONE_MINUTE, ONE_HOUR, ONE_DAY)
    limiters = (HOUR_LIMITER,)

    def __init__(self, name, wall_time_func=time.time, prefix="",
                 bucket_size=ONE_DAY, dsn=None, redis_host=None,
                 redis_port=None, longterm_distinct_days=60):
        """Construct a metric we expect to be periodically updated.

        @param name: Indicates what is being instrumented.
        @param wall_time_func: Function for obtaining wall time.
        @param prefix: If present, a string to prepend to the message
            composed when C{report} is called.
        """
        self.name = name
        self.wall_time_func = wall_time_func
        if prefix:
            prefix += '.'
        self.prefix = prefix
        self.bucket_size = bucket_size
        self.dsn = dsn
        self.redis_host = redis_host
        if redis_port is None:
            redis_port = 6379
        self.redis_port = redis_port
        self.metric_id = None
        self.build_bucket()
        self.redis_flush_lock = threading.Lock()
        self.redis_count = {}
        self.longterm_distinct_count = None
        self.longterm_distinct_days = longterm_distinct_days

        if redis_host is not None:
            self.redis = redis.client.Redis(host=redis_host, port=redis_port)

        if self.dsn is not None:
            reactor.callInThread(self._update_longterm_distinct)

    def build_bucket(self, timestamp=None):
        self.max = 0
        self.bucket = {}
        self.bucket_no = self.get_bucket_no(timestamp)

    def get_bucket_no(self, timestamp=None):
        if timestamp is None:
            timestamp = self.wall_time_func()
        return int(timestamp / (self.bucket_size))

    def getResource(self):
        """Return an http resource to represent this."""
        return DistinctResource(self)

    def process(self, fields):
        self.update(fields[0])

    def update(self, item):
        value = self.bucket.get(item, 0) + 1

        self.bucket[item] = value
        if value > self.max:
            self.max = value

        now = self.wall_time_func()
        if self.redis_host is not None:
            reactor.callInThread(self._update_count, item, now)

    def bucket_name_for(self, period):
        return self.name + "_bucket_" + str(period)

    def _update_count(self, value, when):
        for period in self.periods:
            self.redis.zadd(self.bucket_name_for(period), value, when)

    def _flush_redis(self, now):
        if self.redis_flush_lock.acquire(False) is False:
            return
        try:
            for period in self.periods:
                bucket = self.bucket_name_for(period)
                self.redis.zremrangebyscore(bucket, 0, now - period)
                self.redis_count[bucket] = self.redis.zcard(bucket)
            for item, count in self.bucket.iteritems():
                for limiter in self.limiters:
                    limiter.add(self.redis, "%s:%s" % (self.name, item), count)
        finally:
            self.redis_flush_lock.release()

    def count(self, period):
        return self.redis_count.get(self.bucket_name_for(period), 0)

    def count_5min(self):
        return self.count(5 * ONE_MINUTE)

    def count_1hour(self):
        return self.count(ONE_HOUR)

    def count_1day(self):
        return self.count(ONE_DAY)

    def save_bucket(self, bucket, bucket_no):
        if self.dsn is not None:
            reactor.callInThread(self._save_bucket, bucket, bucket_no)

    def flush(self, interval, timestamp):
        current_bucket = self.get_bucket_no(timestamp)
        if current_bucket != self.bucket_no:
            self.save_bucket(self.bucket, self.bucket_no)
            self.build_bucket(timestamp)

        if self.redis_host is not None:
            reactor.callInThread(self._flush_redis, timestamp)

        metrics = []
        items = {".messages": len(self.bucket),
                 ".max": self.max,
                 ".count_5min": self.count_5min(),
                 ".count_1hour": self.count_1hour(),
                 ".count_1day": self.count_1day(),
                 }

        if self.longterm_distinct_count is not None:
            items[".count_%dday" % (self.longterm_distinct_days,)] = \
                self.longterm_distinct_count

        for item, value in items.iteritems():
            metrics.append((self.prefix + self.name + item, value, timestamp))
        return metrics

    def _save_bucket(self, bucket, bucket_no):
        path = self.prefix + self.name
        c = psycopg2.connect(self.dsn)
        cr = c.cursor()
        if self.metric_id is None:

            cr.execute("SELECT * FROM paths WHERE path = %s", (path,))
            row = cr.fetchone()
            if row is None:
                cr.execute("INSERT INTO paths (path) VALUES (%s) "
                           "RETURNING (id)", (path,))
                row = cr.fetchone()
                cr.execute("commit")

            self.metric_id = row[0]

        for i, (k, v) in enumerate(bucket.iteritems()):
            cr.execute("INSERT INTO points (path_id, bucket, value, count) "
                       "VALUES (%s, %s, %s, %s)", (self.metric_id, bucket_no,
                                                   k, v))
            if i % 1000 == 0:
                cr.execute("commit")
        cr.execute("commit")
        self._update_longterm_distinct()

    def _update_longterm_distinct(self):
        now = self.wall_time_func()
        self.longterm_distinct_count = self._get_distinct_count(
            now - self.longterm_distinct_days * ONE_DAY, now)

    def _get_distinct_count(self, since, until):
        """Get a distinct count for a path between certain ranges."""
        since_bucket = self.get_bucket_no(since)
        until_bucket = self.get_bucket_no(until)
        path = self.prefix + self.name
        c = psycopg2.connect(self.dsn)
        cr = c.cursor()
        cr.execute("SELECT COUNT(DISTINCT value) FROM points "
                   "INNER JOIN paths ON (paths.id = points.path_id) "
                   "WHERE paths.path = %s AND bucket BETWEEN %s AND %s", (
                       path, since_bucket, until_bucket,))
        row = cr.fetchone()
        return row[0]

    def _get_distinct_top_value(self, since, until, how_many=20):
        """Get the top distinct values for a path between certain ranges."""
        since_bucket = self.get_bucket_no(since)
        until_bucket = self.get_bucket_no(until)
        path = self.prefix + self.name
        c = psycopg2.connect(self.dsn)
        cr = c.cursor()
        cr.execute("SELECT value, SUM(count) AS cnt FROM points "
                   "INNER JOIN paths ON (paths.id = points.path_id) "
                   "WHERE paths.path = %s AND bucket BETWEEN %s AND %s"
                   "GROUP BY value ORDER BY cnt DESC, value ASC LIMIT %s", (
                       path, since_bucket, until_bucket, how_many,))
        rows = cr.fetchall()
        return rows
