# Copyright (C) 2011 Canonical
# All Rights Reserved

import json
import ConfigParser
from cStringIO import StringIO
import os
import time

try:
    from subprocess import check_output
except ImportError:
    import subprocess

    def check_output(args):
        return subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]

import psycopg2
import redis

from twisted.trial.unittest import TestCase
from twisted.internet import reactor, protocol, defer
from twisted.plugin import getPlugins
from twisted.plugins import distinctdbplugin
from twisted.web.test.test_web import DummyRequest

from twisted.application import internet
from twisted.web import server
from twisted.web.client import Agent

from txstatsd.itxstatsd import IMetricFactory
from txstatsd import service

from distinctdb import distinctmetric as distinct


class DummyReporter(object):

    def __init__(self):
        self.called = []

    def get_foo(self):
        self.called.append(("get_foo", ()))
        return "foo"

    def _get_distinct_count(self, since, until):
        self.called.append(("get_distinct_count", (since, until)))
        return 42

    def _get_distinct_top_value(self, since, until, how_many=20):
        self.called.append(("get_distinct_top_value",
                            (since, until, how_many)))
        return [("one", 1), ("two", 1)]


class TestDistinctMetricReporter(TestCase):

    def test_get_bucket_no(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)
        self.assertEquals(dmr.get_bucket_no(), 0)
        _wall_time = [60 * 60 * 24 + 1]
        dmr.update("one")
        self.assertEquals(dmr.get_bucket_no(), 1)

    def test_max(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)
        self.assertEquals(dmr.get_bucket_no(), 0)
        self.assertEquals(dmr.max, 0)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        self.assertEquals(dmr.max, 2)
        dmr.flush(1, 60 * 60 * 24 + 1)
        dmr.update("one")
        self.assertEquals(dmr.max, 1)

    def test_reports(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        result = {}

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)

        def save(b, b_no):
            result["bucket"] = b
            result["bucket_no"] = b_no
        dmr.save_bucket = save
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        day = 60 * 60 * 24 + 1
        dmr.flush(1, day)
        dmr.update("three")
        self.assertEquals(result,
                          {"bucket": {"one": 2, "two": 1}, "bucket_no": 0})
        result = dmr.flush(1, day)

        self.assertTrue(("test.max", 1, day) in result)
        self.assertTrue(("test.messages", 1, day) in result)

    def test_configure(self):
        class TestOptions(service.OptionsGlue):
            optParameters = [["test", "t", "default", "help"]]
            config_section = "statsd"

        o = TestOptions()
        config_file = ConfigParser.RawConfigParser()
        config_file.readfp(StringIO("[statsd]\n\n[plugin_distinctdb]\n"
                                    "dsn = dbdsn\nbucket_size = 100\n"
                                    "longterm_distinct_days = 30"))
        o.configure(config_file)
        dmf = distinctdbplugin.DistinctMetricFactory()
        dmf.configure(o)
        dmr = dmf.build_metric("foo", "bar", time.time)
        self.assertEquals(dmr.bucket_size, 100)
        self.assertEquals(dmr.dsn, "dbdsn")
        self.assertEquals(dmr.longterm_distinct_days, 30)


class TestJSONMethodResource(TestCase):

    def test_render_result_as_json(self):
        reporter = DummyReporter()
        request = DummyRequest([])
        resource = distinct.JSONMethodResource(reporter, "get_foo")

        def check(result):
            self.assertEquals({"result": "foo"},
                              json.loads("".join(request.written)))
        d = request.notifyFinish()
        d.addCallback(check)
        d.addErrback(self.fail)
        request.render(resource)
        return d


class TestDistinctResource(TestCase):

    def test_render_top_resource(self):
        reporter = DummyReporter()
        request = DummyRequest([])
        request.args = {"since": [str(time.time())],
                        "until": [str(time.time() + 1)]}
        resource = distinct.DistinctResource(reporter)
        child_resource = resource.getChildWithDefault("top", request)

        def check(result):
            self.assertEquals(json.dumps({"result": [("one", 1), ("two", 1)]}),
                              "".join(request.written))
        d = request.notifyFinish()
        d.addCallback(check)
        request.render(child_resource)
        return d

    def test_render_count_resource(self):
        reporter = DummyReporter()
        request = DummyRequest([])
        request.args = {"since": [str(time.time())],
                        "until": [str(time.time() + 1)]}
        resource = distinct.DistinctResource(reporter)
        child_resource = resource.getChildWithDefault("count", request)

        def check(result):
            self.assertEquals(json.dumps({"result": 42}),
                              "".join(request.written))
        d = request.notifyFinish()
        d.addCallback(check)
        request.render(child_resource)
        return d


class TestPlugin(TestCase):

    def test_factory(self):
        self.assertTrue(distinctdbplugin.distinct_metric_factory in
                        list(getPlugins(IMetricFactory)))


class TestDatabase(TestCase):

    def setUp(self):
        rootdir = check_output(["bzr", "root"]).strip()
        dsn_file = os.path.join(rootdir, "tmp", "pg.dsn")
        self.dsn = open(dsn_file).read()
        self.conn = psycopg2.connect(self.dsn)

    def tearDown(self):
        cr = self.conn.cursor()
        cr.execute("rollback")
        cr.execute("DELETE FROM paths")
        cr.execute("DELETE FROM points")
        cr.execute("commit")

    def _create_test_data_points(self, name, bucket_no):
        """Easily create test data."""
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update(name)
        dmr._save_bucket(dmr.bucket, bucket_no)


class ResponseCollector(protocol.Protocol):

    def __init__(self, finished):
        self.finished = finished
        self.data = []

    def dataReceived(self, bytes):
        self.data.append(bytes)

    def connectionLost(self, reason):
        self.finished.callback("".join(self.data))


def collect_response(response):
    d = defer.Deferred()
    c = ResponseCollector(d)
    response.deliverBody(c)
    return d


class ErrorDummyReporter(object):

    def _get_distinct_count(self, *args, **kweargs):
        raise Exception("die!")


class TestDistinctResourceIntegrationError(TestCase):

    def setUp(self):
        dmr = ErrorDummyReporter()
        root = distinct.DistinctResource(dmr)
        site = server.Site(root)
        self.port = 12341
        self.service = internet.TCPServer(self.port, site)
        self.service.startService()

    @defer.inlineCallbacks
    def test_count(self):
        agent = Agent(reactor)

        result = yield agent.request('GET',
            'http://localhost:%s/count?since=0&until=1000000'
            % (self.port,))
        self.assertEquals(500, result.code)

    def tearDown(self):
        self.service.stopService()


class TestDistinctResourceIntegration(TestDatabase):

    def setUp(self):
        TestDatabase.setUp(self)
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        root = dmr.getResource()
        site = server.Site(root)
        self.port = 12341
        self.service = internet.TCPServer(self.port, site)
        self.service.startService()

    def setup_bucket(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update("one")
        dmr.update("two")
        dmr._save_bucket(dmr.bucket, 1)
        dmr.bucket = {}
        dmr.update("one")
        dmr.update("one")
        dmr._save_bucket(dmr.bucket, 2)

    @defer.inlineCallbacks
    def test_count(self):
        self.setup_bucket()

        agent = Agent(reactor)

        result = yield agent.request('GET',
            'http://localhost:%s/count?since=0&until=1000000'
            % (self.port,))
        self.assertEquals(200, result.code)
        data = yield collect_response(result)
        self.assertEquals({"result": 2}, json.loads(data))

    @defer.inlineCallbacks
    def test_top(self):
        self.setup_bucket()

        agent = Agent(reactor)

        result = yield agent.request('GET',
            'http://localhost:%s/top?since=0&until=1000000&how_many=1'
            % (self.port,))
        self.assertEquals(200, result.code)
        data = yield collect_response(result)
        self.assertEquals({"result": [['one', 3]]}, json.loads(data))

    def tearDown(self):
        self.service.stopService()
        TestDatabase.tearDown(self)


class TestDatabaseMetricStorage(TestDatabase):

    def test_connect(self):
        cr = self.conn.cursor()
        cr.execute("SELECT 0")
        result = cr.fetchall()
        self.assertTrue(result, [(0,)])

    def test_create_metric_id(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr._save_bucket({}, 0)
        cr = self.conn.cursor()
        cr.execute("SELECT * FROM paths WHERE path = 'test'")
        cr.execute("SELECT * FROM paths")
        self.assertEquals(len(cr.fetchall()), 1)

    def test_find_saved_data(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        dmr._save_bucket(dmr.bucket, 0)
        cr = self.conn.cursor()
        cr.execute("SELECT * FROM points ORDER BY value")
        rows = cr.fetchall()
        self.assertEquals(rows, [(dmr.metric_id, 0, "one", 2),
                                 (dmr.metric_id, 0, "two", 1)])

    def test_save_bucket_twice(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        dmr._save_bucket(dmr.bucket, 0)
        dmr.bucket = {}
        dmr.update("one")
        dmr.update("two")
        dmr.update("three")
        dmr._save_bucket(dmr.bucket, 1)
        cr = self.conn.cursor()
        cr.execute("SELECT * FROM points ORDER BY bucket, value")
        rows = cr.fetchall()
        self.assertEquals(rows, [(dmr.metric_id, 0, "one", 2),
                                 (dmr.metric_id, 0, "two", 1),
                                 (dmr.metric_id, 1, "one", 1),
                                 (dmr.metric_id, 1, "three", 1),
                                 (dmr.metric_id, 1, "two", 1)])

    def test_load_metric_id(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr._save_bucket({}, 0)
        dmr2 = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr2._save_bucket({}, 0)
        self.assertEquals(dmr.metric_id, dmr2.metric_id)

    def test_get_distinct_count(self):
        """Test get_distinct_count."""
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        t1 = time.time()
        # add a day
        t2 = time.time() + 60 * 60 * 24 + 1
        b1 = dmr.get_bucket_no(t1)
        b2 = dmr.get_bucket_no(t2)
        # Make sure we're using different buckets
        self.assertNotEqual(b1, b2)
        self._create_test_data_points("one", b1)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 1)
        self._create_test_data_points("two", b1)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 2)
        # Add it again, make sure the distinct is filtering it out
        self._create_test_data_points("two", b1)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 2)
        # Now we add to a newer bucket, but still within range
        self._create_test_data_points("three", b2)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 3)
        # Now to a older bucket, out of range
        self._create_test_data_points("zero", b1 - 1)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 3)
        # Now to a newer bucket, out of range
        self._create_test_data_points("infinity", b2 + 1)
        count = dmr._get_distinct_count(t1, t2)
        self.assertEqual(count, 3)

    def test_get_distinct_top_value(self):
        """Test get_distinct_top_value."""
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        t1 = time.time()
        # add a day
        t2 = time.time() + 60 * 60 * 24 + 1
        b1 = dmr.get_bucket_no(t1)
        b2 = dmr.get_bucket_no(t2)
        # Make sure we're using different buckets
        self.assertNotEqual(b1, b2)
        self._create_test_data_points("one", b1)
        values = dmr._get_distinct_top_value(t1, t2)
        self.assertEqual(values, [("one", 1)])
        self._create_test_data_points("two", b1)
        values = dmr._get_distinct_top_value(t1, t2)
        self.assertEqual(values, [("one", 1), ("two", 1)])
        self._create_test_data_points("one", b1)
        values = dmr._get_distinct_top_value(t1, t2)
        self.assertEqual(values, [("one", 2), ("two", 1)])
        # Create a third, and make "two" have 2 values so "three" is last
        self._create_test_data_points("two", b1)
        self._create_test_data_points("three", b1)
        values = dmr._get_distinct_top_value(t1, t2)
        self.assertEqual(values, [("one", 2), ("two", 2), ("three", 1)])
        # Only get the top 2
        values = dmr._get_distinct_top_value(t1, t2, how_many=2)
        self.assertEqual(values, [("one", 2), ("two", 2)])

    def test_count_long_periods(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        dmr._save_bucket(dmr.bucket, dmr.get_bucket_no() - 2)
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        # this should be running in bg now too
        # but run this way we ensure it has finished before the assert
        dmr._update_longterm_distinct()
        # value is loaded on startup
        self.assertEquals(dmr.longterm_distinct_count, 2)
        dmr.update("one")
        dmr.update("three")
        dmr._save_bucket(dmr.bucket, dmr.get_bucket_no() - 1)
        # value is updated on save_bucket
        self.assertEquals(dmr.longterm_distinct_count, 3)
        # reset bucket
        dmr.build_bucket()
        now = dmr.wall_time_func()
        result = dmr.flush(0, now)
        self.assertTrue(("test.count_60day", 3, now) in result, result)


class TestRedis(TestCase):

    def setUp(self):
        reactor._initThreadPool()
        reactor.threadpool.start()

    def tearDown(self):
        r = redis.client.Redis(host="localhost", port=16379)
        r.flushdb()
        reactor.threadpool.stop()

    def test_connect(self):
        r = redis.client.Redis(host="localhost", port=16379)
        r.ping()

    def test_configure(self):
        class TestOptions(service.OptionsGlue):
            optParameters = [["test", "t", "default", "help"]]
            config_section = "statsd"

        o = TestOptions()
        config_file = ConfigParser.RawConfigParser()
        config_file.readfp(
            StringIO("[statsd]\n\n[plugin_distinctdb]\n"
                     "redis_host = localhost\nredis_port = 16379"))
        o.configure(config_file)
        dmf = distinctdbplugin.DistinctMetricFactory()
        dmf.configure(o)
        dmr = dmf.build_metric("foo", "bar", time.time)
        self.assertEquals(dmr.redis_host, "localhost")
        self.assertEquals(dmr.redis_port, 16379)

    def test_usage(self):
        dmr = distinct.DistinctMetricReporter(
            "test", redis_host="localhost", redis_port=16379)

        self.assertEquals(dmr.count_1hour(), 0)
        dmr._update_count("one", 0)
        dmr._flush_redis(1)
        self.assertEquals(dmr.count_1hour(), 1)
        dmr._update_count("one", 0)
        dmr._flush_redis(1)
        self.assertEquals(dmr.count_1hour(), 1)
        dmr._update_count("two", 30 * distinct.ONE_MINUTE)
        dmr._flush_redis(30 * distinct.ONE_MINUTE)
        self.assertEquals(dmr.count_1hour(), 2)
        dmr._flush_redis(distinct.ONE_HOUR + 10 * distinct.ONE_MINUTE)
        self.assertEquals(dmr.count_1hour(), 1)

    def test_load(self):
        dmr = distinct.DistinctMetricReporter(
            "test", redis_host="localhost", redis_port=16379)
        start = time.time()
        for i in range(10000):
            dmr.update(str(i % 1000))

        while True:
            w = len(reactor.threadpool.working)
            if w == 0:
                break
            time.sleep(0.1)

        dmr._flush_redis(time.time())
        duration = time.time() - start
        self.assertEquals(dmr.count_1hour(), 1000)
        self.assertTrue(duration < 10)
        # self.assertEquals(10, distinct.MINUTE_LIMITER.count(
        #     dmr.redis, "test:42", 10))
        self.assertEquals(10, distinct.HOUR_LIMITER.count(
            dmr.redis, "test:42", 10 * distinct.ONE_MINUTE))
        # self.assertEquals(10, distinct.DAY_LIMITER.count(
        #     dmr.redis, "test:42", 3 * distinct.ONE_HOUR))

    def test_bucket_name(self):
        dmr1 = distinct.DistinctMetricReporter("somename")
        dmr2 = distinct.DistinctMetricReporter("othename")
        dmr3 = distinct.DistinctMetricReporter("somename")
        self.assertEquals(dmr1.bucket_name_for(1), dmr3.bucket_name_for(1))
        self.assertNotEquals(dmr1.bucket_name_for(1), dmr2.bucket_name_for(1))
