import time
import redis

from distinctdb.ratelimit import RateLimiter
from twisted.trial.unittest import TestCase


class Clock(object):

    def __init__(self):
        self.value = 0

    def set(self, value):
        self.value = value

    def __call__(self):
        return self.value


class RateLimitTest(TestCase):

    def setUp(self):
        self.client = redis.client.Redis(host="localhost", port=16379)
        self.clock = Clock()
        self.limiter = RateLimiter(span=60, bucket_size=10,
                                   tick=self.clock)
        return super(RateLimitTest, self).setUp()

    def tearDown(self):
        self.client.flushdb()
        return super(RateLimitTest, self).tearDown()

    def test_count(self):
        for i in xrange(10):
            self.clock.set(i * 10)
            self.limiter.add(self.client, 'test')

        self.assertEqual(6, self.limiter.count(self.client, 'test', 80))
        self.assertEqual(6, self.limiter.count(self.client, 'test', 60))
        self.assertEqual(4, self.limiter.count(self.client, 'test', 40))
        self.assertEqual(2, self.limiter.count(self.client, 'test', 20))
        self.assertEqual(1, self.limiter.count(self.client, 'test', 10))

        # Asking for an interval smaller than the bucket size rounds it up to
        # one bucket. We do this to avoid getting false negatives when
        # accidentally asking for too small an interval.
        self.assertEqual(1, self.limiter.count(self.client, 'test', 1))

        self.clock.set(100)
        self.assertEqual(3, self.limiter.count(self.client, 'test', 40))
        self.assertEqual(1, self.limiter.count(self.client, 'test', 20))
        self.assertEqual(0, self.limiter.count(self.client, 'test', 10))

        self.clock.set(110)
        self.assertEqual(2, self.limiter.count(self.client, 'test', 40))
        self.assertEqual(0, self.limiter.count(self.client, 'test', 20))
        self.assertEqual(0, self.limiter.count(self.client, 'test', 10))


class RateLimitExpireTest(TestCase):

    def setUp(self):
        self.client = redis.client.Redis(host="localhost", port=16379)
        self.clock = Clock()
        self.limiter = RateLimiter(span=1, bucket_size=1,
                                   tick=self.clock)
        return super(RateLimitExpireTest, self).setUp()

    def tearDown(self):
        self.client.flushdb()
        return super(RateLimitExpireTest, self).tearDown()

    def test_expire(self):
        for i in xrange(2):
            self.limiter.add(self.client, 'test')

        self.assertEqual(2, self.limiter.count(self.client, 'test', 1))
        time.sleep(2)
        self.assertEqual(0, self.limiter.count(self.client, 'test', 1))
